@@ -49,12 +49,17 @@ class CompletionRefresher(object):
 
     def _bg_refresh(self, pgexecute, special, callbacks, history=None,
       settings=None):
+        settings = settings or {}
         completer = PGCompleter(smart_completion=True, pgspecial=special,
             settings=settings)
 
-        # Create a new pgexecute method to popoulate the completions.
-        e = pgexecute
-        executor = PGExecute(e.dbname, e.user, e.password, e.host, e.port, e.dsn)
+        if settings.get('single_connection'):
+            executor = pgexecute
+        else:
+            # Create a new pgexecute method to popoulate the completions.
+            e = pgexecute
+            executor = PGExecute(
+                e.dbname, e.user, e.password, e.host, e.port, e.dsn)
 
         # If callbacks is a single function then push it into a list.
         if callable(callbacks):
