@@ -11,7 +11,7 @@ import threading
 import shutil
 import functools
 import humanize
-from time import time
+from time import time, sleep
 from codecs import open
 
 
@@ -84,6 +84,8 @@ class NullHandler(logging.Handler):
 
 class PGCli(object):
 
+    default_prompt = '\\u@\\h:\\d> '
+
     def set_default_pager(self, config):
         configured_pager = config['main'].get('pager')
         os_environ_pager = os.environ.get('PAGER')
@@ -101,7 +103,8 @@ class PGCli(object):
         os.environ['LESS'] = '-SRXF'
 
     def __init__(self, force_passwd_prompt=False, never_passwd_prompt=False,
-                 pgexecute=None, pgclirc_file=None, row_limit=None):
+                 pgexecute=None, pgclirc_file=None, row_limit=None,
+                 single_connection=False):
 
         self.force_passwd_prompt = force_passwd_prompt
         self.never_passwd_prompt = never_passwd_prompt
@@ -118,6 +121,7 @@ class PGCli(object):
         self.pgspecial = PGSpecial()
 
         self.multi_line = c['main'].as_bool('multi_line')
+        self.multiline_mode = c['main'].get('multi_line_mode', 'psql')
         self.vi_mode = c['main'].as_bool('vi')
         self.pgspecial.timing_enabled = c['main'].as_bool('timing')
         if row_limit is not None:
@@ -130,20 +134,29 @@ class PGCli(object):
         self.cli_style = c['colors']
         self.wider_completion_menu = c['main'].as_bool('wider_completion_menu')
         self.less_chatty = c['main'].as_bool('less_chatty')
-
+        self.null_string = c['main'].get('null_string', '<null>')
+        self.prompt_format = c['main'].get('prompt', self.default_prompt)
         self.on_error = c['main']['on_error'].upper()
-
+        self.decimal_format = c['data_formats']['decimal']
+        self.float_format = c['data_formats']['float']
         self.completion_refresher = CompletionRefresher()
 
         self.query_history = []
 
         # Initialize completer
         smart_completion = c['main'].as_bool('smart_completion')
-        settings = {'casing_file': get_casing_file(c),
+        keyword_casing = c['main']['keyword_casing']
+        self.settings = {
+            'casing_file': get_casing_file(c),
             'generate_casing_file': c['main'].as_bool('generate_casing_file'),
-            'asterisk_column_order': c['main']['asterisk_column_order']}
+            'generate_aliases': c['main'].as_bool('generate_aliases'),
+            'asterisk_column_order': c['main']['asterisk_column_order'],
+            'single_connection': single_connection,
+            'keyword_casing': keyword_casing,
+        }
+
         completer = PGCompleter(smart_completion, pgspecial=self.pgspecial,
-            settings=settings)
+            settings=self.settings)
         self.completer = completer
         self._completer_lock = threading.Lock()
         self.register_special_commands()
@@ -342,6 +355,70 @@ class PGCli(object):
             continue
         return document
 
+    def execute_command(self, text, query):
+        logger = self.logger
+
+        try:
+            output, query = self._evaluate_command(text)
+        except KeyboardInterrupt:
+            # Restart connection to the database
+            self.pgexecute.connect()
+            logger.debug("cancelled query, sql: %r", text)
+            click.secho("cancelled query", err=True, fg='red')
+        except NotImplementedError:
+            click.secho('Not Yet Implemented.', fg="yellow")
+        except OperationalError as e:
+            if ('server closed the connection'
+                    in utf8tounicode(e.args[0])):
+                self._handle_server_closed_connection()
+            else:
+                logger.error("sql: %r, error: %r", text, e)
+                logger.error("traceback: %r", traceback.format_exc())
+                click.secho(str(e), err=True, fg='red')
+        except Exception as e:
+            logger.error("sql: %r, error: %r", text, e)
+            logger.error("traceback: %r", traceback.format_exc())
+            click.secho(str(e), err=True, fg='red')
+        else:
+            try:
+                if self.output_file and not text.startswith(('\\o ', '\\? ')):
+                    try:
+                        with open(self.output_file, 'a', encoding='utf-8') as f:
+                            click.echo(text, file=f)
+                            click.echo('\n'.join(output), file=f)
+                            click.echo('', file=f)  # extra newline
+                    except IOError as e:
+                        click.secho(str(e), err=True, fg='red')
+                else:
+                    click.echo_via_pager('\n'.join(output))
+            except KeyboardInterrupt:
+                pass
+
+            if self.pgspecial.timing_enabled:
+                # Only add humanized time display if > 1 second
+                if query.total_time > 1:
+                    print('Time: %0.03fs (%s)' % (query.total_time,
+                          humanize.time.naturaldelta(query.total_time)))
+                else:
+                    print('Time: %0.03fs' % query.total_time)
+
+            # Check if we need to update completions, in order of most
+            # to least drastic changes
+            if query.db_changed:
+                with self._completer_lock:
+                    self.completer.reset_completions()
+                self.refresh_completions(persist_priorities='keywords')
+            elif query.meta_changed:
+                self.refresh_completions(persist_priorities='all')
+            elif query.path_changed:
+                logger.debug('Refreshing search path')
+                with self._completer_lock:
+                    self.completer.set_search_path(
+                        self.pgexecute.search_path())
+                logger.debug('Search path: %r',
+                             self.completer.search_path)
+        return query
+
     def run_cli(self):
         logger = self.logger
 
@@ -382,65 +459,17 @@ class PGCli(object):
                 # Initialize default metaquery in case execution fails
                 query = MetaQuery(query=document.text, successful=False)
 
-                try:
-                    output, query = self._evaluate_command(document.text)
-                except KeyboardInterrupt:
-                    # Restart connection to the database
-                    self.pgexecute.connect()
-                    logger.debug("cancelled query, sql: %r", document.text)
-                    click.secho("cancelled query", err=True, fg='red')
-                except NotImplementedError:
-                    click.secho('Not Yet Implemented.', fg="yellow")
-                except OperationalError as e:
-                    if ('server closed the connection'
-                            in utf8tounicode(e.args[0])):
-                        self._handle_server_closed_connection()
-                    else:
-                        logger.error("sql: %r, error: %r", document.text, e)
-                        logger.error("traceback: %r", traceback.format_exc())
-                        click.secho(str(e), err=True, fg='red')
-                except Exception as e:
-                    logger.error("sql: %r, error: %r", document.text, e)
-                    logger.error("traceback: %r", traceback.format_exc())
-                    click.secho(str(e), err=True, fg='red')
+                watch_command, timing = special.get_watch_command(document.text)
+                if watch_command:
+                    while watch_command:
+                        try:
+                            query = self.execute_command(watch_command, query)
+                            click.echo('Waiting for {0} seconds before repeating'.format(timing))
+                            sleep(timing)
+                        except KeyboardInterrupt:
+                            watch_command = None
                 else:
-                    try:
-                        if self.output_file and not document.text.startswith(('\\o ', '\\? ')):
-                            try:
-                                with open(self.output_file, 'a', encoding='utf-8') as f:
-                                    click.echo(document.text, file=f)
-                                    click.echo('\n'.join(output), file=f)
-                                    click.echo('', file=f) # extra newline
-                            except IOError as e:
-                                click.secho(str(e), err=True, fg='red')
-                        else:
-                            click.echo_via_pager('\n'.join(output))
-                    except KeyboardInterrupt:
-                        pass
-
-                    if self.pgspecial.timing_enabled:
-                        # Only add humanized time display if > 1 second
-                        if query.total_time > 1:
-                            print('Time: %0.03fs (%s)' % (query.total_time,
-                                  humanize.time.naturaldelta(query.total_time)))
-                        else:
-                            print('Time: %0.03fs' % query.total_time)
-
-                    # Check if we need to update completions, in order of most
-                    # to least drastic changes
-                    if query.db_changed:
-                        with self._completer_lock:
-                            self.completer.reset_completions()
-                        self.refresh_completions(persist_priorities='keywords')
-                    elif query.meta_changed:
-                        self.refresh_completions(persist_priorities='all')
-                    elif query.path_changed:
-                        logger.debug('Refreshing search path')
-                        with self._completer_lock:
-                            self.completer.set_search_path(
-                                self.pgexecute.search_path())
-                        logger.debug('Search path: %r',
-                                     self.completer.search_path)
+                    query = self.execute_command(document.text, query)
 
                 # Allow PGCompleter to learn user's preferred keywords, etc.
                 with self._completer_lock:
@@ -462,7 +491,7 @@ class PGCli(object):
             set_vi_mode_enabled=set_vi_mode)
 
         def prompt_tokens(_):
-            return [(Token.Prompt, '%s> ' % self.pgexecute.dbname)]
+            return [(Token.Prompt, self.get_prompt(self.prompt_format))]
 
         def get_continuation_tokens(cli, width):
             return [(Token.Continuation, '.' * (width - 1) + ' ')]
@@ -488,6 +517,7 @@ class PGCli(object):
         with self._completer_lock:
             buf = PGBuffer(
                 always_multiline=self.multi_line,
+                multiline_mode=self.multiline_mode,
                 completer=self.completer,
                 history=history,
                 complete_while_typing=Always(),
@@ -557,12 +587,11 @@ class PGCli(object):
                 max_width = None
 
             formatted = format_output(
-                title, cur, headers, status, self.table_format,
-                self.pgspecial.expanded_output, max_width)
+                title, cur, headers, status, self.table_format, self.decimal_format,
+                self.float_format, self.null_string,  self.pgspecial.expanded_output, max_width)
 
             output.extend(formatted)
-            end = time()
-            total += end - start
+            total = time() - start
 
             # Keep track of whether any of the queries are mutating or changing
             # the database
@@ -602,12 +631,8 @@ class PGCli(object):
 
         callback = functools.partial(self._on_completions_refreshed,
                                      persist_priorities=persist_priorities)
-        c = self.config
-        settings = {'casing_file': get_casing_file(c),
-          'generate_casing_file': c['main'].as_bool('generate_casing_file'),
-          'asterisk_column_order': c['main']['asterisk_column_order']}
         self.completion_refresher.refresh(self.pgexecute, self.pgspecial,
-            callback, history=history, settings=settings)
+            callback, history=history, settings=self.settings)
         return [(None, None, None,
                 'Auto-completion refresh started in the background.')]
 
@@ -660,6 +685,16 @@ class PGCli(object):
             return self.completer.get_completions(
                 Document(text=text, cursor_position=cursor_positition), None)
 
+    def get_prompt(self, string):
+        string = string.replace('\\u', self.pgexecute.user or '(none)')
+        string = string.replace('\\h', self.pgexecute.host or '(none)')
+        string = string.replace('\\d', self.pgexecute.dbname or '(none)')
+        string = string.replace('\\p', str(self.pgexecute.port) or '(none)')
+        string = string.replace('\\i', str(self.pgexecute.pid) or '(none)')
+        string = string.replace('\\#', "#" if (self.pgexecute.superuser) else ">")
+        string = string.replace('\\n', "\n")
+        return string
+
 
 @click.command()
 # Default host is '' so psycopg2 can default to either localhost or unix socket
@@ -673,6 +708,9 @@ class PGCli(object):
         help='Force password prompt.')
 @click.option('-w', '--no-password', 'never_prompt', is_flag=True,
         default=False, help='Never prompt for password.')
+@click.option('--single-connection', 'single_connection', is_flag=True,
+        default=False,
+        help='Do not use a separate connection for completions.')
 @click.option('-v', '--version', is_flag=True, help='Version of pgcli.')
 @click.option('-d', '--dbname', default='', envvar='PGDATABASE',
         help='database name to connect to.')
@@ -684,8 +722,8 @@ class PGCli(object):
         help='Set threshold for row limit prompt. Use 0 to disable prompt.')
 @click.argument('database', default=lambda: None, envvar='PGDATABASE', nargs=1)
 @click.argument('username', default=lambda: None, envvar='PGUSER', nargs=1)
-def cli(database, user, host, port, prompt_passwd, never_prompt, dbname,
-        username, version, pgclirc, dsn, row_limit):
+def cli(database, user, host, port, prompt_passwd, never_prompt,
+    single_connection, dbname, username, version, pgclirc, dsn, row_limit):
 
     if version:
         print('Version:', __version__)
@@ -708,7 +746,7 @@ def cli(database, user, host, port, prompt_passwd, never_prompt, dbname,
                    config_full_path)
 
     pgcli = PGCli(prompt_passwd, never_prompt, pgclirc_file=pgclirc,
-                  row_limit=row_limit)
+                  row_limit=row_limit, single_connection=single_connection)
 
     # Choose which ever one has a valid value.
     database = database or dbname
@@ -754,21 +792,23 @@ def obfuscate_process_password():
 
     setproctitle.setproctitle(process_title)
 
-def format_output(title, cur, headers, status, table_format, expanded=False, max_width=None):
+
+def format_output(title, cur, headers, status, table_format, dcmlfmt, floatfmt,
+                  missingval='<null>', expanded=False, max_width=None):
     output = []
     if title:  # Only print the title if it's not None.
         output.append(title)
     if cur:
         headers = [utf8tounicode(x) for x in headers]
         if expanded and headers:
-            output.append(expanded_table(cur, headers))
+            output.append(expanded_table(cur, headers, missingval))
         else:
             tabulated, rows = tabulate(cur, headers, tablefmt=table_format,
-                missingval='<null>')
+                missingval=missingval, dcmlfmt=dcmlfmt, floatfmt=floatfmt)
             if (max_width and rows and
                     content_exceeds_width(rows[0], max_width) and
                     headers):
-                output.append(expanded_table(rows, headers))
+                output.append(expanded_table(rows, headers, missingval))
             else:
                 output.append(tabulated)
     if status:  # Only print the status if it's not None.
