@@ -4,8 +4,10 @@ import re
 import sqlparse
 from collections import namedtuple
 from sqlparse.sql import Comparison, Identifier, Where
-from .parseutils import (
-    last_word, extract_tables, find_prev_keyword, parse_partial_identifier)
+from .parseutils.utils import (
+    last_word, find_prev_keyword, parse_partial_identifier)
+from .parseutils.tables import extract_tables
+from .parseutils.ctes import isolate_query_ctes
 from pgspecial.main import parse_special_command
 
 PY2 = sys.version_info[0] == 2
@@ -20,20 +22,28 @@ else:
 Special = namedtuple('Special', [])
 Database = namedtuple('Database', [])
 Schema = namedtuple('Schema', [])
-Table = namedtuple('Table', ['schema'])
+# FromClauseItem is a table/view/function used in the FROM clause
+# `table_refs` contains the list of tables/... already in the statement,
+# used to ensure that the alias we suggest is unique
+FromClauseItem = namedtuple('FromClauseItem', 'schema table_refs local_tables')
+Table = namedtuple('Table', ['schema', 'table_refs', 'local_tables'])
+View = namedtuple('View', ['schema', 'table_refs'])
 # JoinConditions are suggested after ON, e.g. 'foo.barid = bar.barid'
-JoinCondition = namedtuple('JoinCondition', ['tables', 'parent'])
+JoinCondition = namedtuple('JoinCondition', ['table_refs', 'parent'])
 # Joins are suggested after JOIN, e.g. 'foo ON foo.barid = bar.barid'
-Join = namedtuple('Join', ['tables', 'schema'])
+Join = namedtuple('Join', ['table_refs', 'schema'])
 
-Function = namedtuple('Function', ['schema', 'filter'])
+Function = namedtuple('Function', ['schema', 'table_refs', 'filter'])
 # For convenience, don't require the `filter` argument in Function constructor
-Function.__new__.__defaults__ = (None, None)
+Function.__new__.__defaults__ = (None, tuple(), None)
+Table.__new__.__defaults__ = (None, tuple(), tuple())
+View.__new__.__defaults__ = (None, tuple())
+FromClauseItem.__new__.__defaults__ = (None, tuple(), tuple())
 
-Column = namedtuple('Column', ['tables', 'require_last_table'])
-Column.__new__.__defaults__ = (None, None)
+Column = namedtuple(
+    'Column', ['table_refs', 'require_last_table', 'local_tables'])
+Column.__new__.__defaults__ = (None, None, tuple())
 
-View = namedtuple('View', ['schema'])
 Keyword = namedtuple('Keyword', [])
 NamedQuery = namedtuple('NamedQuery', [])
 Datatype = namedtuple('Datatype', ['schema'])
@@ -49,6 +59,10 @@ class SqlStatement(object):
             text_before_cursor, include='many_punctuations')
         full_text = _strip_named_query(full_text)
         text_before_cursor = _strip_named_query(text_before_cursor)
+
+        full_text, text_before_cursor, self.local_tables = \
+            isolate_query_ctes(full_text, text_before_cursor)
+
         self.text_before_cursor_including_last_word = text_before_cursor
 
         # If we've partially typed a word then word_before_cursor won't be an
@@ -74,7 +88,25 @@ class SqlStatement(object):
         self.text_before_cursor = text_before_cursor
         self.parsed = parsed
 
-        self.last_token = parsed and parsed.token_prev(len(parsed.tokens)) or ''
+        self.last_token = parsed and parsed.token_prev(len(parsed.tokens))[1] or ''
+
+    def is_insert(self):
+        return self.parsed.token_first().value.lower() == 'insert'
+
+    def get_tables(self, scope='full'):
+        """ Gets the tables available in the statement.
+        param `scope:` possible values: 'full', 'insert', 'before'
+        If 'insert', only the first table is returned.
+        If 'before', only tables before the cursor are returned.
+        If not 'insert' and the stmt is an insert, the first table is skipped.
+        """
+        tables = extract_tables(
+            self.full_text if scope == 'full' else self.text_before_cursor)
+        if scope == 'insert':
+            tables = tables[:1]
+        elif self.is_insert():
+            tables = tables[1:]
+        return tables
 
     def get_identifier_schema(self):
         schema = (self.identifier and self.identifier.get_parent_name()) or None
@@ -84,9 +116,9 @@ class SqlStatement(object):
 
         return schema
 
-    def reduce_to_prev_keyword(self):
+    def reduce_to_prev_keyword(self, n_skip=0):
         prev_keyword, self.text_before_cursor = \
-            find_prev_keyword(self.text_before_cursor)
+            find_prev_keyword(self.text_before_cursor, n_skip=n_skip)
         return prev_keyword
 
 
@@ -101,7 +133,12 @@ def suggest_type(full_text, text_before_cursor):
     if full_text.startswith('\\i '):
         return (Path(),)
 
-    stmt = SqlStatement(full_text, text_before_cursor)
+    # This is a temporary hack; the exception handling
+    # here should be removed once sqlparse has been fixed
+    try:
+        stmt = SqlStatement(full_text, text_before_cursor)
+    except (TypeError, AttributeError):
+        return []
 
     # Check for special commands and handle those separately
     if stmt.parsed:
@@ -139,7 +176,7 @@ def _split_multiple_statements(full_text, text_before_cursor, parsed):
         stmt_start, stmt_end = 0, 0
 
         for statement in parsed:
-            stmt_len = len(statement.to_unicode())
+            stmt_len = len(str(statement))
             stmt_start, stmt_end = stmt_end, stmt_end + stmt_len
 
             if stmt_end >= current_pos:
@@ -267,7 +304,7 @@ def suggest_based_on_last_token(token, stmt):
 
             # Check for a subquery expression (cases 3 & 4)
             where = p.tokens[-1]
-            prev_tok = where.token_prev(len(where.tokens) - 1)
+            prev_tok = where.token_prev(len(where.tokens) - 1)[1]
 
             if isinstance(prev_tok, Comparison):
                 # e.g. "SELECT foo FROM bar WHERE foo = ANY("
@@ -280,15 +317,17 @@ def suggest_based_on_last_token(token, stmt):
                 return column_suggestions
 
         # Get the token before the parens
-        prev_tok = p.token_prev(len(p.tokens) - 1)
+        prev_tok = p.token_prev(len(p.tokens) - 1)[1]
 
         if (prev_tok and prev_tok.value
           and prev_tok.value.lower().split(' ')[-1] == 'using'):
             # tbl1 INNER JOIN tbl2 USING (col1, col2)
-            tables = extract_tables(stmt.text_before_cursor)
+            tables = stmt.get_tables('before')
 
             # suggest columns that are present in more than one table
-            return (Column(tables=tables, require_last_table=True),)
+            return (Column(table_refs=tables,
+                           require_last_table=True,
+                           local_tables=stmt.local_tables),)
 
         elif p.token_first().value.lower() == 'select':
             # If the lparen is preceeded by a space chances are we're about to
@@ -296,51 +335,59 @@ def suggest_based_on_last_token(token, stmt):
             if last_word(stmt.text_before_cursor,
                          'all_punctuations').startswith('('):
                 return (Keyword(),)
+        prev_prev_tok = prev_tok and p.token_prev(p.token_index(prev_tok))[1]
+        if prev_prev_tok and prev_prev_tok.normalized == 'INTO':
+            return (Column(table_refs=stmt.get_tables('insert')),)
         # We're probably in a function argument list
-        return (Column(tables=extract_tables(stmt.full_text)),)
+        return (Column(table_refs=extract_tables(stmt.full_text),
+                       local_tables=stmt.local_tables),)
     elif token_v in ('set', 'by', 'distinct'):
-        return (Column(tables=extract_tables(stmt.full_text)),)
+        return (Column(table_refs=stmt.get_tables(),
+                       local_tables=stmt.local_tables),)
     elif token_v in ('select', 'where', 'having'):
         # Check for a table alias or schema qualification
         parent = (stmt.identifier and stmt.identifier.get_parent_name()) or []
-
+        tables = stmt.get_tables()
         if parent:
-            tables = extract_tables(stmt.full_text)
             tables = tuple(t for t in tables if identifies(parent, t))
-            return (Column(tables=tables),
+            return (Column(table_refs=tables, local_tables=stmt.local_tables),
                     Table(schema=parent),
                     View(schema=parent),
                     Function(schema=parent),)
         else:
-            return (Column(tables=extract_tables(stmt.full_text)),
+            return (Column(table_refs=tables, local_tables=stmt.local_tables),
                     Function(schema=None),
                     Keyword(),)
-
+    elif token_v == 'as':
+        # Don't suggest anything for aliases
+        return ()
     elif (token_v.endswith('join') and token.is_keyword) or (token_v in
             ('copy', 'from', 'update', 'into', 'describe', 'truncate')):
 
         schema = stmt.get_identifier_schema()
+        tables = extract_tables(stmt.text_before_cursor)
+        is_join = token_v.endswith('join') and token.is_keyword
 
         # Suggest tables from either the currently-selected schema or the
         # public schema if no schema has been specified
-        suggest = [Table(schema=schema)]
+        suggest = []
 
         if not schema:
             # Suggest schemas
             suggest.insert(0, Schema())
 
-        # Only tables can be TRUNCATED, otherwise suggest views
-        if token_v != 'truncate':
-            suggest.append(View(schema=schema))
+        if token_v == 'from' or is_join:
+            suggest.append(FromClauseItem(schema=schema,
+                                          table_refs=tables,
+                                          local_tables=stmt.local_tables))
+        elif token_v == 'truncate':
+            suggest.append(Table(schema))
+        else:
+            suggest.extend((Table(schema), View(schema)))
 
-        # Suggest set-returning functions in the FROM clause
-        if token_v == 'from' or (token_v.endswith('join') and token.is_keyword):
-            suggest.append(Function(schema=schema, filter='for_from_clause'))
-
-        if (token_v.endswith('join') and token.is_keyword
-          and _allow_join(stmt.parsed)):
-            tables = extract_tables(stmt.text_before_cursor)
-            suggest.append(Join(tables=tables, schema=schema))
+        if is_join and _allow_join(stmt.parsed):
+            tables = stmt.get_tables('before')
+            suggest.append(Join(table_refs=tables, schema=schema))
 
         return tuple(suggest)
 
@@ -355,21 +402,22 @@ def suggest_based_on_last_token(token, stmt):
 
     elif token_v == 'column':
         # E.g. 'ALTER TABLE foo ALTER COLUMN bar
-        return (Column(tables=extract_tables(stmt.text_before_cursor)),)
+        return (Column(table_refs=stmt.get_tables()),)
 
     elif token_v == 'on':
-        tables = extract_tables(stmt.text_before_cursor)  # [(schema, table, alias), ...]
+        tables = stmt.get_tables('before')
         parent = (stmt.identifier and stmt.identifier.get_parent_name()) or None
         if parent:
             # "ON parent.<suggestion>"
             # parent can be either a schema name or table alias
             filteredtables = tuple(t for t in tables if identifies(parent, t))
-            sugs = [Column(tables=filteredtables),
+            sugs = [Column(table_refs=filteredtables,
+                           local_tables=stmt.local_tables),
                     Table(schema=parent),
                     View(schema=parent),
                     Function(schema=parent)]
             if filteredtables and _allow_join_condition(stmt.parsed):
-                sugs.append(JoinCondition(tables=tables,
+                sugs.append(JoinCondition(table_refs=tables,
                                           parent=filteredtables[-1]))
             return tuple(sugs)
         else:
@@ -378,7 +426,7 @@ def suggest_based_on_last_token(token, stmt):
             aliases = tuple(t.ref for t in tables)
             if _allow_join_condition(stmt.parsed):
                 return (Alias(aliases=aliases), JoinCondition(
-                    tables=tables, parent=None))
+                    table_refs=tables, parent=None))
             else:
                 return (Alias(aliases=aliases),)
 
@@ -406,6 +454,14 @@ def suggest_based_on_last_token(token, stmt):
         if not schema:
             suggestions.append(Schema())
         return tuple(suggestions)
+    elif token.is_keyword:
+        # token is a keyword we haven't implemented any special handling for
+        # go backwards in the query until we find one we do recognize
+        prev_keyword = stmt.reduce_to_prev_keyword(n_skip=1)
+        if prev_keyword:
+            return suggest_based_on_last_token(prev_keyword, stmt)
+        else:
+            return (Keyword(),)
     else:
         return (Keyword(),)
 
@@ -433,7 +489,7 @@ def _allow_join_condition(statement):
     if not statement or not statement.tokens:
         return False
 
-    last_tok = statement.token_prev(len(statement.tokens))
+    last_tok = statement.token_prev(len(statement.tokens))[1]
     return last_tok.value.lower() in ('on', 'and', 'or')
 
 
@@ -452,6 +508,6 @@ def _allow_join(statement):
     if not statement or not statement.tokens:
         return False
 
-    last_tok = statement.token_prev(len(statement.tokens))
+    last_tok = statement.token_prev(len(statement.tokens))[1]
     return (last_tok.value.lower().endswith('join')
         and last_tok.value.lower() not in('cross join', 'natural join'))
