@@ -1,21 +1,22 @@
 from __future__ import print_function, unicode_literals
 import logging
 import re
-import itertools
+from itertools import count, repeat, chain
 import operator
 from collections import namedtuple, defaultdict
 from pgspecial.namedqueries import NamedQueries
 from prompt_toolkit.completion import Completer, Completion
 from prompt_toolkit.contrib.completers import PathCompleter
 from prompt_toolkit.document import Document
-from .packages.sqlcompletion import (
+from .packages.sqlcompletion import (FromClauseItem,
     suggest_type, Special, Database, Schema, Table, Function, Column, View,
     Keyword, NamedQuery, Datatype, Alias, Path, JoinCondition, Join)
-from .packages.function_metadata import ColumnMetadata, ForeignKey
-from .packages.parseutils import last_word, TableReference
+from .packages.parseutils.meta import ColumnMetadata, ForeignKey
+from .packages.parseutils.utils import last_word
+from .packages.parseutils.tables import TableReference
 from .packages.pgliterals.main import get_literals
 from .packages.prioritization import PrevalenceCounter
-from .config import load_config, config_location, get_config
+from .config import load_config, config_location
 
 try:
     from collections import OrderedDict
@@ -32,6 +33,16 @@ Match = namedtuple('Match', ['completion', 'priority'])
 
 normalize_ref = lambda ref: ref if ref[0] == '"' else '"' + ref.lower() +  '"'
 
+def generate_alias(tbl, tbs):
+    """ Generate a table alias, consisting of all upper-case letters in
+    the table name, or, if there are no upper-case letters, the first letter +
+    all letters preceded by _
+    param tbl - unescaped name of the table to alias
+    param tbls - set TableReference objects for tables already in query
+    """
+    return ''.join([l for l in tbl if l.isupper()] or
+        [l for l, prev in zip(tbl,  '_' + tbl) if prev == '_' and l != '_'])
+
 class PGCompleter(Completer):
     keywords = get_literals('keywords')
     functions = get_literals('functions')
@@ -43,11 +54,17 @@ class PGCompleter(Completer):
         self.pgspecial = pgspecial
         self.prioritizer = PrevalenceCounter()
         settings = settings or {}
+        self.generate_aliases = settings.get('generate_aliases')
         self.casing_file = settings.get('casing_file')
         self.generate_casing_file = settings.get('generate_casing_file')
         self.asterisk_column_order = settings.get(
             'asterisk_column_order', 'table_order')
 
+        keyword_casing = settings.get('keyword_casing', 'upper').lower()
+        if keyword_casing not in ('upper', 'lower', 'auto'):
+            keyword_casing = 'upper'
+        self.keyword_casing = keyword_casing
+
         self.reserved_words = set()
         for x in self.keywords:
             self.reserved_words.update(x.split())
@@ -222,7 +239,7 @@ class PGCompleter(Completer):
 
     def find_matches(self, text, collection, mode='fuzzy',
                      meta=None, meta_collection=None,
-                     type_priority=0, priority_collection = None):
+                     priority_collection = None):
         """Find completion matches for the given text.
 
         Given the user's input text and a collection of available
@@ -238,6 +255,11 @@ class PGCompleter(Completer):
 
         """
 
+        prio_order = [
+            'keyword', 'function', 'view', 'table', 'datatype', 'database',
+            'schema', 'column', 'table alias', 'join', 'name join', 'fk join'
+        ]
+        type_priority = prio_order.index(meta) if meta in prio_order else -1
         text = last_word(text, include='most_punctuations').lower()
         text_len = len(text)
 
@@ -279,9 +301,9 @@ class PGCompleter(Completer):
                     return -float('Infinity'), -match_point
 
         # Fallback to meta param if meta_collection param is None
-        meta_collection = meta_collection or itertools.repeat(meta)
+        meta_collection = meta_collection or repeat(meta)
         # Fallback to 0 if priority_collection param is None
-        priority_collection = priority_collection or itertools.repeat(0)
+        priority_collection = priority_collection or repeat(0)
 
         collection = zip(collection, meta_collection, priority_collection)
 
@@ -299,12 +321,16 @@ class PGCompleter(Completer):
                 # position. Since we use *higher* priority to mean "more
                 # important," we use -ord(c) to prioritize "aa" > "ab" and end
                 # with 1 to prioritize shorter strings (ie "user" > "users").
+                # We first do a case-insensitive sort and then a
+                # case-sensitive one as a tie breaker.
                 # We also use the unescape_name to make sure quoted names have
                 # the same priority as unquoted names.
-                lexical_priority = tuple(-ord(c) for c in self.unescape_name(item)) + (1,)
+                lexical_priority = (tuple(0 if c in(' _') else -ord(c)
+                    for c in self.unescape_name(item.lower())) + (1,)
+                    + tuple(c for c in item))
 
                 item = self.case(item)
-                priority = type_priority, prio, sort_key, priority_func(item), lexical_priority
+                priority = sort_key, type_priority, prio, priority_func(item), lexical_priority
 
                 matches.append(Match(
                     completion=Completion(item, -text_len, display_meta=meta),
@@ -346,12 +372,14 @@ class PGCompleter(Completer):
 
         return [m.completion for m in matches]
 
+
     def get_column_matches(self, suggestion, word_before_cursor):
-        tables = suggestion.tables
+        tables = suggestion.table_refs
         _logger.debug("Completion column scope: %r", tables)
-        scoped_cols = self.populate_scoped_cols(tables)
+        scoped_cols = self.populate_scoped_cols(tables, suggestion.local_tables)
+
         colit = scoped_cols.items
-        flat_cols = list(itertools.chain(*((c.name for c in cols)
+        flat_cols = list(chain(*((c.name for c in cols)
             for t, cols in colit())))
         if suggestion.require_last_table:
             # require_last_table is used for 'tb11 JOIN tbl2 USING (...' which should
@@ -383,21 +411,28 @@ class PGCompleter(Completer):
             return [Match(completion=Completion(collist, -1,
                 display_meta='columns', display='*'), priority=(1,1,1))]
 
-        return self.find_matches(word_before_cursor, flat_cols, meta='column')
+        return self.find_matches(word_before_cursor, flat_cols,
+            meta='column')
 
-    def generate_alias(self, tbl, tbls):
+    def alias(self, tbl, tbls):
         """ Generate a unique table alias
         tbl - name of the table to alias, quoted if it needs to be
-        tbls - set of table refs already in use, normalized with normalize_ref
+        tbls - TableReference iterable of tables already in query
         """
-        if tbl[0] == '"':
-            aliases = ('"' + tbl[1:-1] + str(i) + '"' for i in itertools.count(2))
+        tbl = self.case(tbl)
+        tbls = set(normalize_ref(t.ref) for t in tbls)
+        if self.generate_aliases:
+            tbl = generate_alias(self.unescape_name(tbl), tbls)
+        if normalize_ref(tbl) not in tbls:
+            return tbl
+        elif tbl[0] == '"':
+            aliases = ('"' + tbl[1:-1] + str(i) + '"' for i in count(2))
         else:
-            aliases = (self.case(tbl) + str(i) for i in itertools.count(2))
+            aliases = (tbl + str(i) for i in count(2))
         return next(a for a in aliases if normalize_ref(a) not in tbls)
 
     def get_join_matches(self, suggestion, word_before_cursor):
-        tbls = suggestion.tables
+        tbls = suggestion.table_refs
         cols = self.populate_scoped_cols(tbls)
         # Set up some data structures for efficient access
         qualified = dict((normalize_ref(t.ref), t.schema) for t in tbls)
@@ -418,8 +453,8 @@ class PGCompleter(Completer):
             if suggestion.schema and left.schema != suggestion.schema:
                 continue
             c = self.case
-            if normalize_ref(left.tbl) in refs:
-                lref = self.generate_alias(left.tbl, refs)
+            if self.generate_aliases or normalize_ref(left.tbl) in refs:
+                lref = self.alias(left.tbl, suggestion.table_refs)
                 join = '{0} {4} ON {4}.{1} = {2}.{3}'.format(
                     c(left.tbl), c(left.col), rtbl.ref, c(right.col), lref)
             else:
@@ -436,26 +471,27 @@ class PGCompleter(Completer):
                 0 if (left.schema, left.tbl) in other_tbls else 1))
 
         return self.find_matches(word_before_cursor, joins, meta='join',
-            priority_collection=prios, type_priority=100)
+            priority_collection=prios)
 
     def get_join_condition_matches(self, suggestion, word_before_cursor):
         col = namedtuple('col', 'schema tbl col')
-        tbls = self.populate_scoped_cols(suggestion.tables).items
+        tbls = self.populate_scoped_cols(suggestion.table_refs).items
         cols = [(t, c) for t, cs in tbls() for c in cs]
         try:
-            lref = (suggestion.parent or suggestion.tables[-1]).ref
+            lref = (suggestion.parent or suggestion.table_refs[-1]).ref
             ltbl, lcols = [(t, cs) for (t, cs) in tbls() if t.ref == lref][-1]
         except IndexError: # The user typed an incorrect table qualifier
             return []
-        conds, found_conds = [], set()
+        conds, prios, found_conds = [], [], set()
 
-        def add_cond(lcol, rcol, rref, meta, prio):
+        def add_cond(lcol, rcol, rref, prio):
             prefix = '' if suggestion.parent else ltbl.ref + '.'
             case = self.case
             cond = prefix + case(lcol) + ' = ' + rref + '.' + case(rcol)
             if cond not in found_conds:
                 found_conds.add(cond)
-                conds.append((cond, meta, prio + ref_prio[rref]))
+                conds.append(cond)
+                prios.append(prio + ref_prio[rref])
 
         def list_dict(pairs): # Turns [(a, b), (a, c)] into {a: [b, c]}
             d = defaultdict(list)
@@ -465,7 +501,7 @@ class PGCompleter(Completer):
 
         # Tables that are closer to the cursor get higher prio
         ref_prio = dict((tbl.ref, num) for num, tbl
-            in enumerate(suggestion.tables))
+            in enumerate(suggestion.table_refs))
         # Map (schema, table, col) to tables
         coldict = list_dict(((t.schema, t.name, c.name), t)
             for t, c in cols if t.ref != lref)
@@ -478,29 +514,35 @@ class PGCompleter(Completer):
             par = col(fk.parentschema, fk.parenttable, fk.parentcolumn)
             left, right = (child, par) if left == child else (par, child)
             for rtbl in coldict[right]:
-                add_cond(left.col, right.col, rtbl.ref, 'fk join', 2000)
+                add_cond(left.col, right.col, rtbl.ref, 2000)
+        matches = self.find_matches(word_before_cursor, conds,
+          meta='fk join', priority_collection=prios)
         # For name matching, use a {(colname, coltype): TableReference} dict
         coltyp = namedtuple('coltyp', 'name datatype')
         col_table = list_dict((coltyp(c.name, c.datatype), t) for t, c in cols)
+        conds, prios = [], []
         # Find all name-match join conditions
         for c in (coltyp(c.name, c.datatype) for c in lcols):
             for rtbl in (t for t in col_table[c] if t.ref != ltbl.ref):
-                add_cond(c.name, c.name, rtbl.ref, 'name join', 1000
+                add_cond(c.name, c.name, rtbl.ref, 1000
                     if c.datatype in ('integer', 'bigint', 'smallint') else 0)
 
-        conds, metas, prios = zip(*conds) if conds else ([], [], [])
+        return matches + self.find_matches(word_before_cursor, conds,
+          meta='name join', priority_collection=prios)
 
-        return self.find_matches(word_before_cursor, conds,
-          meta_collection=metas, type_priority=100, priority_collection=prios)
-
-    def get_function_matches(self, suggestion, word_before_cursor):
+    def get_function_matches(self, suggestion, word_before_cursor, alias=False):
         if suggestion.filter == 'for_from_clause':
             # Only suggest functions allowed in FROM clause
             filt = lambda f: not f.is_aggregate and not f.is_window
             funcs = self.populate_functions(suggestion.schema, filt)
+            if alias:
+                funcs = [self.case(f) + '() ' + self.alias(f,
+                    suggestion.table_refs) for f in funcs]
+            else:
+                funcs = [self.case(f) + '()' for f in funcs]
         else:
-            funcs = self.populate_schema_objects(
-                suggestion.schema, 'functions')
+            funcs = [f + '()' for f in self.populate_schema_objects(
+                suggestion.schema, 'functions')]
 
         # Function overloading means we way have multiple functions of the same
         # name at this point, so keep unique names only
@@ -525,26 +567,44 @@ class PGCompleter(Completer):
         if not word_before_cursor.startswith('pg_'):
             schema_names = [s for s in schema_names if not s.startswith('pg_')]
 
-        return self.find_matches(word_before_cursor, schema_names, meta='schema')
+        return self.find_matches(word_before_cursor, schema_names,
+            meta='schema')
 
-    def get_table_matches(self, suggestion, word_before_cursor):
+    def get_from_clause_item_matches(self, suggestion, word_before_cursor):
+        alias = self.generate_aliases
+        s = suggestion
+        t_sug = Table(s.schema, s.table_refs, s.local_tables)
+        v_sug = View(s.schema, s.table_refs)
+        f_sug = Function(s.schema, s.table_refs, filter='for_from_clause')
+        return (self.get_table_matches(t_sug, word_before_cursor, alias)
+            + self.get_view_matches(v_sug, word_before_cursor, alias)
+            + self.get_function_matches(f_sug, word_before_cursor, alias))
+
+    def get_table_matches(self, suggestion, word_before_cursor, alias=False):
         tables = self.populate_schema_objects(suggestion.schema, 'tables')
+        tables.extend(tbl.name for tbl in suggestion.local_tables)
 
         # Unless we're sure the user really wants them, don't suggest the
         # pg_catalog tables that are implicitly on the search path
         if not suggestion.schema and (
                 not word_before_cursor.startswith('pg_')):
             tables = [t for t in tables if not t.startswith('pg_')]
+        if alias:
+            tables = [self.case(t) + ' ' + self.alias(t, suggestion.table_refs)
+                for t in tables]
+        return self.find_matches(word_before_cursor, tables,
+            meta='table')
 
-        return self.find_matches(word_before_cursor, tables, meta='table')
 
-    def get_view_matches(self, suggestion, word_before_cursor):
+    def get_view_matches(self, suggestion, word_before_cursor, alias=False):
         views = self.populate_schema_objects(suggestion.schema, 'views')
 
         if not suggestion.schema and (
                 not word_before_cursor.startswith('pg_')):
             views = [v for v in views if not v.startswith('pg_')]
-
+        if alias:
+            views = [self.case(v) + ' ' + self.alias(v, suggestion.table_refs)
+                for v in views]
         return self.find_matches(word_before_cursor, views, meta='view')
 
     def get_alias_matches(self, suggestion, word_before_cursor):
@@ -557,7 +617,19 @@ class PGCompleter(Completer):
                                  meta='database')
 
     def get_keyword_matches(self, _, word_before_cursor):
-        return self.find_matches(word_before_cursor, self.keywords,
+        casing = self.keyword_casing
+        if casing == 'auto':
+            if word_before_cursor and word_before_cursor[-1].islower():
+                casing = 'lower'
+            else:
+                casing = 'upper'
+
+        if casing == 'upper':
+            keywords = [k.upper() for k in self.keywords]
+        else:
+            keywords = [k.lower() for k in self.keywords]
+
+        return self.find_matches(word_before_cursor, keywords,
                                  mode='strict', meta='keyword')
 
     def get_path_matches(self, _, word_before_cursor):
@@ -594,6 +666,7 @@ class PGCompleter(Completer):
             word_before_cursor, NamedQueries.instance.list(), meta='named query')
 
     suggestion_matchers = {
+        FromClauseItem: get_from_clause_item_matches,
         JoinCondition: get_join_condition_matches,
         Join: get_join_matches,
         Column: get_column_matches,
@@ -610,9 +683,10 @@ class PGCompleter(Completer):
         Path: get_path_matches,
     }
 
-    def populate_scoped_cols(self, scoped_tbls):
+    def populate_scoped_cols(self, scoped_tbls, local_tbls=()):
         """ Find all columns in a set of scoped_tables
         :param scoped_tbls: list of TableReference namedtuples
+        :param local_tbls: tuple(TableMetadata)
         :return: {TableReference:{colname:ColumnMetaData}}
         """
 
@@ -645,6 +719,10 @@ class PGCompleter(Completer):
                             cols = cols.values()
                             addcols(schema, relname, tbl.alias, reltype, cols)
                             break
+                            
+        # Local tables should shadow database tables
+        for tbl in local_tbls:
+            columns[tbl.name] = tbl.columns
 
         return columns
 
@@ -663,7 +741,7 @@ class PGCompleter(Completer):
             objects = [obj for schema in schemas
                            for obj in metadata[schema].keys()]
 
-        return objects
+        return [self.case(o) for o in objects]
 
     def populate_functions(self, schema, filter_func):
         """Returns a list of function names
