@@ -1,15 +1,15 @@
 import traceback
 import logging
-import itertools
 import psycopg2
 import psycopg2.extras
 import psycopg2.extensions as ext
 import sqlparse
 import pgspecial as special
-from .packages.function_metadata import FunctionMetadata, ForeignKey
+import select
+from psycopg2.extensions import POLL_OK, POLL_READ, POLL_WRITE
+from .packages.parseutils.meta import FunctionMetadata, ForeignKey
 from .encodingutils import unicode2utf8, PY2, utf8tounicode
 
-
 _logger = logging.getLogger(__name__)
 
 # Cast all database input to unicode automatically.
@@ -25,9 +25,36 @@ ext.register_type(ext.new_type((2249,), "RECORD", ext.UNICODE))
 # Postgres 9+ and as escaped binary in earlier versions.
 ext.register_type(ext.new_type((17,), 'BYTEA_TEXT', psycopg2.STRING))
 
+# TODO: Get default timeout from pgclirc?
+_WAIT_SELECT_TIMEOUT = 1
+
+
+def _wait_select(conn):
+    """
+        copy-pasted from psycopg2.extras.wait_select
+        the default implementation doesn't define a timeout in the select calls
+    """
+    while 1:
+        try:
+            state = conn.poll()
+            if state == POLL_OK:
+                break
+            elif state == POLL_READ:
+                select.select([conn.fileno()], [], [], _WAIT_SELECT_TIMEOUT)
+            elif state == POLL_WRITE:
+                select.select([], [conn.fileno()], [], _WAIT_SELECT_TIMEOUT)
+            else:
+                raise conn.OperationalError("bad state from poll: %s" % state)
+        except KeyboardInterrupt:
+            conn.cancel()
+            # the loop will be broken by a server error
+            continue
+
+
 # When running a query, make pressing CTRL+C raise a KeyboardInterrupt
 # See http://initd.org/psycopg/articles/2014/07/20/cancelling-postgresql-statements-python/
-ext.set_wait_callback(psycopg2.extras.wait_select)
+# See also https://github.com/psycopg/psycopg2/issues/468
+ext.set_wait_callback(_wait_select)
 
 
 def register_date_typecasters(connection):
@@ -42,7 +69,7 @@ def register_date_typecasters(connection):
     date_oid = cursor.description[0][1]
     cursor.execute('SELECT NULL::timestamp')
     timestamp_oid = cursor.description[0][1]
-    cursor.execute('SELECT NULL::timestamptz')
+    cursor.execute('SELECT NULL::timestamp with time zone')
     timestamptz_oid = cursor.description[0][1]
     oids = (date_oid, timestamp_oid, timestamptz_oid)
     new_type = psycopg2.extensions.new_type(oids, 'DATE', cast_date)
@@ -136,6 +163,7 @@ class PGExecute(object):
         host = (host or self.host)
         port = (port or self.port)
         dsn = (dsn or self.dsn)
+        pid = -1
         if dsn:
             if password:
                 dsn = "{0} password={1}".format(dsn, password)
@@ -143,10 +171,10 @@ class PGExecute(object):
             cursor = conn.cursor()
             # When we connect using a DSN, we don't really know what db,
             # user, etc. we connected to. Let's read it.
-            db = self._select_one(cursor, 'select current_database()')
-            user = self._select_one(cursor, 'select current_user')
-            host = self._select_one(cursor, 'select inet_server_addr()')
-            port = self._select_one(cursor, 'select inet_server_port()')
+            db = self._select_one(cursor, 'select current_database()')[0]
+            user = self._select_one(cursor, 'select current_user')[0]
+            host = self._select_one(cursor, 'select inet_server_addr()')[0]
+            port = self._select_one(cursor, 'select inet_server_port()')[0]
         else:
             conn = psycopg2.connect(
                     database=unicode2utf8(db),
@@ -154,6 +182,13 @@ class PGExecute(object):
                     password=unicode2utf8(password),
                     host=unicode2utf8(host),
                     port=unicode2utf8(port))
+
+            cursor = conn.cursor()
+
+        cursor.execute("SHOW ALL")
+        db_parameters = dict(name_val_desc[:2] for name_val_desc in cursor.fetchall())
+
+        pid = self._select_one(cursor, 'select pg_backend_pid()')[0]
         conn.set_client_encoding('utf8')
         if hasattr(self, 'conn'):
             self.conn.close()
@@ -164,6 +199,8 @@ class PGExecute(object):
         self.password = password
         self.host = host
         self.port = port
+        self.pid = pid
+        self.superuser = db_parameters.get('is_superuser') == '1'
 
         register_date_typecasters(conn)
         register_json_typecasters(self.conn, self._json_typecaster)
